
class State(object):
    INITIALIZED = 'initialized'
    ESTOP = 'estop'
    BYPASSED = 'bypassed'
    IDLE = 'idle'
    UNLOCKED = 'unlocked'
    IN_USE = 'in_use'
    IN_TRAINING = 'in_training'
    TERMINATED = 'terminated'
