# noinspection PyClassHasNoInit
class ClientOption:
    DEBUG = 'debug'
    PID_FILE = 'pid_file'
    LOG_FILE = 'log_file'
    LOG_LEVEL = 'log_level'
    DEVICE_ID = 'device_id'
    PIN_ESTOP = 'pin_estop'
    USE_ESTOP = 'use_estop'
    PIN_ALARM = 'pin_alarm'
    USE_ALARM = 'use_alarm'
    IS_A_DOOR = 'is_a_door'
    PIN_LOGOUT = 'pin_logout'
    PIN_3V3_EN = 'pin_3v3_en'
    USE_3V3_EN = 'use_3v3_en'
    CONFIG_FILE = 'config_file'
    PIN_LED_RED = 'pin_led_red'
    AUTO_UPDATE = 'auto_update'
    STATUS_FILE = 'status_file'
    PIN_LED_BLUE = 'pin_led_blue'
    REBOOT_DELAY = 'reboot_delay'
    FORCE_UPDATE = 'force_update'
    PIN_LED_GREEN = 'pin_led_green'
    SERVER_ADDRESS = 'server_address'
    DISPLAY_SERLCD = 'display_serlcd'
    CT_TURNS_RATIO = 'ct_turns_ratio'
    ESTOP_ACTIVE_HI = 'estop_active_hi'
    REBOOT_ON_ERROR = 'reboot_on_error'
    PIN_POWER_RELAY = 'pin_power_relay'
    SERIAL_PORT_NAME = 'serial_port_name'
    DOOR_UNLOCK_TIME = 'door_unlock_time'
    SERIAL_PORT_SPEED = 'serial_port_speed'
    PIN_CURRENT_SENSE = 'pin_current_sense'
    PIN_BYPASS_DETECT = 'pin_bypass_detect'
    USE_BYPASS_DETECT = 'use_bypass_detect'
    LOGOUT_COAST_TIME = 'logout_coast_time'
    CT_BURDEN_RESISTOR = 'ct_burden_resistor'
    LOGGING_CONFIG_FILE = 'logging_config_file'
    ALLOW_USER_OVERRIDE = 'allow_user_override'
    AUTO_UPDATE_INTERVAL = 'auto_update_interval'
    PIN_CURRENT_THRESHOLD = 'pin_current_threshold'
    DISABLE_TRAINING_MODE = 'disable_training_mode'
    DOOR_CONTINUOUS_UNLOCK = 'door_continuous_unlock'
    CURRENT_DETECT_SETTING = 'current_detect_setting'
    MAX_POWER_DOWN_TIMEOUT = 'max_power_down_timeout'
    USE_PGM_CURRENT_THRESHOLD = 'use_pgm_current_threshold'
